@@ -4,13 +4,13 @@ use 5.008000;
 use strict;
 use warnings;
 
-our $VERSION = '0.02';
+our $VERSION = '0.04';
 
 use File::Spec;
 use YAML::XS qw( LoadFile );
 use Cpanel::JSON::XS;
 use Hash::Merge;
-use Data::Rmap qw( rmap_to :types );
+use Scalar::Util qw( refaddr );
 use Carp qw( croak );
 
 my %FILE_EXTENSIONS_MAP = (
@@ -57,6 +57,7 @@ sub new {
   $self->{_hash_merge} = Hash::Merge->new( 'CONFIG_PRECEDENT' );
   $self->{_config}     = undef;
   $self->{_vars}       = {};
+  $self->{_seen_nodes} = {};
 
   return $self;
 }
@@ -81,13 +82,11 @@ sub load {
   my $self            = shift;
   my @config_sections = @_;
 
-  $self->{_config} = $self->_build_tree( @config_sections );
+  $self->{_config} = $self->_build_tree(@config_sections);
+  $self->_process_tree( $self->{_config} );
 
-  rmap_to {
-    $_ = $self->_process_node( $_ );
-  } HASH|VALUE, $self->{_config};
-
-  $self->{_vars} = {};
+  $self->{_vars}       = {};
+  $self->{_seen_nodes} = {};
 
   return $self->{_config};
 }
@@ -182,13 +181,42 @@ sub _load_json_file {
   return @data;
 }
 
+sub _process_tree {
+  my $self = shift;
+
+  $_[0] = $self->_process_node( $_[0] );
+
+  if ( my $node_addr = refaddr( $_[0] ) ) {
+    return if $self->{_seen_nodes}{$node_addr};
+
+    $self->{_seen_nodes}{$node_addr} = 1;
+  }
+
+  if ( ref( $_[0] ) eq 'HASH' ) {
+    foreach ( values %{ $_[0] } ) {
+      $self->_process_tree($_);
+    }
+  }
+  elsif ( ref( $_[0] ) eq 'ARRAY' ) {
+    foreach ( @{ $_[0] } ) {
+      $self->_process_tree($_);
+    }
+  }
+
+  return;
+}
+
 sub _process_node {
   my $self = shift;
   my $node = shift;
 
   return unless defined $node;
 
-  if ( ref($node) eq 'HASH' && $self->{process_directives} ) {
+  if ( !ref($node) && $self->{interpolate_variables} ) {
+    $node =~ s/\$((\$?)\{([^\}]*)\})/
+        $2 ? $1 : ( $self->_resolve_var( $3 ) || '' )/ge;
+  }
+  elsif ( ref($node) eq 'HASH' && $self->{process_directives} ) {
     if ( defined $node->{var} ) {
       $node = $self->_resolve_var( $node->{var} );
     }
@@ -209,10 +237,6 @@ sub _process_node {
       }
     }
   }
-  elsif ( $self->{interpolate_variables} ) { # SCALAR
-    $node =~ s/\$((\$?)\{([^\}]*)\})/
-        $2 ? $1 : $self->_resolve_var( $3 )/ge;
-  }
 
   return $node;
 }
@@ -244,10 +268,12 @@ sub _resolve_var {
 
   my $vars = $self->{_vars};
 
-  unless ( defined $vars->{$var_name}  ) {
+  unless ( defined $vars->{$var_name} ) {
     my @tokens = split( /\./, $var_name );
     my $pointer = $self->{_config};
 
+    my $value;
+
     while ( 1 ) {
       my $token = shift @tokens;
 
@@ -255,9 +281,8 @@ sub _resolve_var {
         last unless defined $pointer->{$token};
 
         if ( !@tokens ) {
-          $vars->{$var_name}
-              = $self->_process_node( $pointer->{$token} );
-          $pointer->{$token} = $vars->{$var_name};
+          $value = $self->_process_node( $pointer->{$token} );
+          $pointer->{$token} = $value;
 
           last;
         }
@@ -275,9 +300,8 @@ sub _resolve_var {
         last unless defined $pointer->[$token];
 
         if ( !@tokens ) {
-          $vars->{$var_name}
-              = $self->_process_node( $pointer->[$token] );
-          $pointer->[$token] = $vars->{$var_name};
+          $value = $self->_process_node( $pointer->[$token] );
+          $pointer->[$token] = $value;
 
           last;
         }
@@ -287,10 +311,10 @@ sub _resolve_var {
         $pointer = $pointer->[$token];
       }
     }
-  }
 
-  unless ( defined $vars->{$var_name} ) {
-    $vars->{$var_name} = '';
+    if ( defined $value ) {
+      $vars->{$var_name} = $value;
+    }
   }
 
   return $vars->{$var_name};
@@ -312,7 +336,11 @@ features
     dirs => [ qw( /etc/myapp /home/username/etc/myapp ) ],
   );
 
-  my $config = $config_processor->load( qw( dirs.yml db.json metrics/* ),
+  my $config = $config_processor->load( qw( dirs.yml db.json metrics/* ) );
+
+  $config = $config_processor->load(
+    qw( dirs.yml db.json redis.yml mongodb.json metrics/* ),
+
     { myapp => {
         db => {
           connectors => {
@@ -339,11 +367,13 @@ F<.yaml>, F<.jsn>, F<.json>.
 =head2 new( %params )
 
   my $config_processor = Config::Processor->new(
-    dirs => [ qw( /etc/myapp /home/username/myapp/etc ) ],
+    dirs => [ qw( /etc/myapp /home/username/etc/myapp ) ],
   );
 
+  $config_processor = Config::Processor->new();
+
   $config_processor = Config::Processor->new(
-    dirs                  => [ qw( /etc/myapp /home/username/myapp/etc ) ],
+    dirs                  => [ qw( /etc/myapp /home/username/etc/myapp ) ],
     interpolate_variables => 0,
     process_directives    => 0,
   );
@@ -352,7 +382,8 @@ F<.yaml>, F<.jsn>, F<.json>.
 
 =item dirs => \@dirs
 
-List of directories, in which configuration processor will search files.
+List of directories, in which configuration processor will search files. If
+the parameter not specified, current directory will be used.
 
 =item interpolate_variables => $boolean
 
@@ -368,9 +399,6 @@ Enables or disables directive processing. Enabled by default.
 
 =head2 load( @config_sections )
 
-  my $config = $config_processor->load( qw( dirs.yml db.json metrics/* ),
-      \%local_config );
-
 Attempts to load all configuration sections and returns reference to resulting
 configuration tree.
 
@@ -378,6 +406,8 @@ Configuration section can be a relative filename, a filename with wildcard
 characters or a hash reference. Filenames with wildcard characters is processed
 by C<CORE::glob> function and supports the same syntax.
 
+  my $config = $config_processor->load( qw( myapp.yml extras/* ), \%hard_config );
+
 =head2 interpolate_variables( [ $boolean ] )
 
 Enables or disables variable interpolation in configurations files.
